/*
  gcc -lX11 xtopmost.c
 */
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#define _NET_WM_STATE_REMOVE        0
#define _NET_WM_STATE_ADD           1
#define _NET_WM_STATE_TOGGLE        2

Bool set_always_on_top(Display *dpy, Window wnd)
{
	Atom wmStateAbove, wmNetWmState;
	XClientMessageEvent xclient;

	wmStateAbove = XInternAtom(dpy, "_NET_WM_STATE_ABOVE", 1);
	if (wmStateAbove == None)
		goto fail_above;

	wmNetWmState = XInternAtom(dpy, "_NET_WM_STATE", 1);
	if (wmNetWmState == None)
		goto fail_state;

	memset(&xclient, 0, sizeof(xclient));
	xclient.type = ClientMessage;
	xclient.window = wnd;
	xclient.message_type = wmNetWmState;
	xclient.format = 32;
	xclient.data.l[0] = _NET_WM_STATE_ADD;
	xclient.data.l[1] = wmStateAbove;

	XSendEvent(dpy, DefaultRootWindow(dpy), False,
		   SubstructureRedirectMask | SubstructureNotifyMask,
		   (XEvent *)&xclient);

	XFlush(dpy);
	return True;
fail_above:
	fputs("cannot find _NET_WM_STATE_ABOVE atom\n", stderr);
	return False;
fail_state:
	fputs("cannot find _NET_WM_STATE atom\n", stderr);
	return False;
}

int main(void)
{
	int run = 1, status = EXIT_FAILURE;
	Atom atom_wm_delete;
	Display* dpy;
	Window wnd;
	XEvent e;

	dpy = XOpenDisplay(NULL);
	if (!dpy) {
		fputs("cannot open display!\n", stderr);
		return EXIT_FAILURE;
	}

	atom_wm_delete = XInternAtom(dpy, "WM_DELETE_WINDOW", True);
	if (atom_wm_delete == None) {
		fputs("cannot find WM_DELETE_WINDOW atom\n", stderr);
		return EXIT_FAILURE;
	}

	/* create window */
	wnd = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy),
				  0, 0, 200, 100, 0, 0, 0);
	if (!wnd) {
		fputs("cannot create window!\n", stderr);
		goto outdpy;
	}

	XSelectInput(dpy, wnd, ExposureMask | StructureNotifyMask);
	XSetWMProtocols(dpy, wnd, &atom_wm_delete, 1);
	XFlush(dpy);

	XStoreName(dpy, wnd, "Always on top!");
	XMapWindow(dpy, wnd);

	set_always_on_top(dpy, wnd);

	/* event loop */
	while (run) {
		XNextEvent(dpy, &e);

		switch (e.type) {
		case Expose:
			XClearWindow(dpy, e.xany.window);
			XFlush(dpy);
			break;
		case ClientMessage:
			if (e.xclient.data.l[0] == (long)atom_wm_delete) {
				XUnmapWindow(dpy, wnd);
				run = 0;
			}
			break;
		}
	}

	status = 0;

	/* cleanup */
	status = EXIT_SUCCESS;
	XDestroyWindow(dpy, wnd);
outdpy:
	XCloseDisplay(dpy);
	return status;
}
